# cython: language_level=3
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True

# group structure that can become a cluster
# Author: Pavel Artamonov
# License: 3-clause BSD

import numpy as np
cimport numpy as np

from libc.math cimport fabs, pow

# def allocate_buffer_groups(np.intp_t size, np.intp_t n_dim)

cdef set_precision(np.double_t prec)

cdef class Group (object):
    cdef:
        data
        __data_length
        _size
        _neg_uniq_edges
        assume_data(self, data, s, ue)
        mtn_set_like(self, ogroup)

        points(self)
        uniq_edges(self)
        child(self, np.intp_t c, is_outlier)
        mtn_coords(self)
        mtn_center(self)
        mtn_weight(self)

        aggregate(self, np.double_t v, np.intp_t is_cluster1, Group group1, np.intp_t is_cluster2, Group group2)
        mtn_aggregate(self, v, is_cluster1, group1, is_cluster2, group2)
        mtn_subtract(self, ogroup, v)
        add_1_autocluster(self, np.double_t border)
        cook_outlier(self, np.double_t border)
        mtn_add_1_coords(self, coords, border)
        mtn_set_coords(self, coords)

        will_cluster(self, np.double_t border, np.double_t neg_common_coef)
        mtn_mark_cluster(self, bint is_cluster)
        mtn_need_cluster(self)
        mtn_change_sum_edges(self, v)



# cython: language_level=3
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True

# group structure that can become a cluster
# Author: Pavel Artamonov
# License: 3-clause BSD

import numpy as np
cimport numpy as np

from libc.math cimport fabs, pow

# cdef int _IDX_POINTS = 0 # size has to be first!!!
# cdef int _IDX_UNIQ_EDGES = 1
cdef int _IDX_SUM_RECIPROCALS = 2

cdef int _IDX_UF_HAS_LEFT_CHILD = _IDX_SUM_RECIPROCALS + 1 # for clustering by UF
# cdef int _IDX_DICT_SIZES = 6

cdef int _IDX_MTN_SUM_EDGES = _IDX_UF_HAS_LEFT_CHILD + 1
cdef int _IDX_MTN_WILL_CLUSTER = _IDX_UF_HAS_LEFT_CHILD + 2
cdef int _IDX_MTN_SUM_COORDS = _IDX_UF_HAS_LEFT_CHILD + 3

cdef np.double_t _group_PRECISION = 0.0000001

cdef set_precision(np.double_t prec):
    _group_PRECISION = prec

def allocate_buffer_groups(np.intp_t size, np.intp_t n_dim=0):
    fields = [("points", np.intp),
              ("uniq_edges", np.intp),
              ("sum_reciprocals", np.double),
              ("has_left_child", np.intp),
              # ("dict_points", dict),
     ]
    if n_dim != 0: # for motion
        fields.append(("sum_edges", np.double)),
        fields.append(("will_cluster", np.intp)),
        fields.append(("sum_coords", np.double, n_dim))


    dtype = np.dtype(fields, align=True)
    return np.empty(size, dtype=dtype)

def allocate_buffer_clusters(np.intp_t num_points):
    return np.empty((num_points - 1), dtype=np.intp)

def allocate_buffer_sizes(np.intp_t num_points):
    return np.empty((num_points - 1), dtype=np.intp)

cdef class Group (object):
    # declarations are in pxd file
    # https://cython.readthedocs.io/en/latest/src/userguide/sharing_declarations.html

    def __init__(self, data):
        self.data = data
        self.__data_length = len(data)
        self._size = 0
        self._neg_uniq_edges = 0 # edges are negative until proven clusters


    cdef assume_data(self, data, s, ue):
        self.data = data
        self._size = s
        self._neg_uniq_edges = ue

    cdef points(self):
        # return self.data[_IDX_POINTS]
        return self._size
    cdef uniq_edges(self):
        # return self.data[_IDX_UNIQ_EDGES]
        return self._neg_uniq_edges

    cdef aggregate(self, np.double_t v, np.intp_t is_cluster1, Group group1, np.intp_t is_cluster2, Group group2):
    # cdef aggregate(self, v, is_cluster1, group1, is_cluster2, group2):

        cdef np.intp_t i

        self._size = group1._size + group2._size

        # edges are negative until proven clusters
        self._neg_uniq_edges = -1 + \
                               (0 if is_cluster1 else group1._neg_uniq_edges) + \
                               (0 if is_cluster2 else group2._neg_uniq_edges)

        i = _IDX_SUM_RECIPROCALS
        self.data[i] = (0 if not is_cluster1 and not is_cluster2 else 1./v) \
                       + ((group1.data[i]) if not is_cluster1 else 0) \
                       + ((group2.data[i]) if not is_cluster2 else 0)

    cdef add_1_autocluster(self, np.double_t border):
        if self._size == 1:
            self._size = 2
            self._neg_uniq_edges = -1
            self.data[_IDX_SUM_RECIPROCALS] = 1./border
            return

        self._size += 1
        self._neg_uniq_edges += -1
        self.data[_IDX_SUM_RECIPROCALS] += 1. / border

    cdef cook_outlier(self, np.double_t border):
        self._size = 1
        # self.data[_IDX_UNIQ_EDGES] = 0
        self._neg_uniq_edges = 0
        self.data[_IDX_SUM_RECIPROCALS] = 0

    cdef child(self, np.intp_t c, is_outlier):
        if self.data[_IDX_UF_HAS_LEFT_CHILD] == 0: # setter
            self.data[_IDX_UF_HAS_LEFT_CHILD] = -1 if is_outlier else c+1
            return False, False, 0
        # getter
        return True, self.data[_IDX_UF_HAS_LEFT_CHILD] == -1, self.data[_IDX_UF_HAS_LEFT_CHILD]-1

    cdef will_cluster(self, np.double_t border, np.double_t neg_common_coef):
        cdef bint is_cluster

        new_form = 1. * border * self.data[_IDX_SUM_RECIPROCALS] * self._neg_uniq_edges * neg_common_coef
        old_shells = self._size - 1 + _group_PRECISION
        is_cluster = new_form > old_shells
        # print("{:.2f}".format(border),
        #       'is_cluster', "{:.2f}".format(new_form / old_shells),
        #       new_form > old_shells,
        #       "{:.1f}".format(new_form),
        #       '>', "{:.1f}".format(old_shells),
        #       'N/K', "{:.2f}".format(-neg_common_coef),
        #       'clusters vs size', -self._neg_uniq_edges, '/', self._size,
        #       'avg.border',
        # )

        return is_cluster

# ----------------------------
    cdef mtn_mark_cluster(self, bint is_cluster):
        self.data[_IDX_MTN_WILL_CLUSTER] = is_cluster
    cdef mtn_need_cluster(self):
        return self.data[_IDX_MTN_WILL_CLUSTER]

    cdef mtn_change_sum_edges(self, v):
        self.data[_IDX_MTN_SUM_EDGES] = v * (self._size - 1)
    cdef mtn_weight(self):
        return self.data[_IDX_MTN_SUM_EDGES]

# -----------------------
    cdef mtn_set_like(self, base_group):
        cdef np.intp_t i
        i = self.__data_length
        while i != 0:
            i -= 1
            self.data[i] = base_group[i]

    cdef mtn_subtract(self, ogroup, v):
        i = self.__data_length
        while i!=0:
            i -= 1
            self.data[i] -= ogroup[i]
        self.data[_IDX_MTN_SUM_EDGES] -= v

    cdef mtn_aggregate(self, v, is_cluster1, group1, is_cluster2, group2):
        cdef np.intp_t i

        i = _IDX_MTN_SUM_EDGES
        self.data[i] = v + (((group1[0]-1) * v) if is_cluster1 else group1[i]) + \
                       (((group2[0]-1) * v) if is_cluster2 else group2[i])

        i = _IDX_MTN_SUM_COORDS
        self.data[i] += group1[i] + group2[i]
        #TODO: merge mean?  https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance

#----------for motion purposes------------------
    cdef mtn_coords(self):
        return self.data[_IDX_MTN_SUM_COORDS]
    cdef mtn_set_coords(self, coords):
        self.data[_IDX_MTN_SUM_COORDS] = coords

    cdef mtn_add_1_coords(self, coords, border):
        self.data[_IDX_MTN_SUM_COORDS] += coords
        if self._size == 1:
            self.data[_IDX_MTN_SUM_EDGES] = border

    cdef mtn_center(self):
        return self.data[_IDX_MTN_SUM_COORDS] / self._size

# ------------------- debug

    # cdef aggregate_dict(self, v, is_cluster1, group1, is_cluster2, group2):
    #
    #     i = _IDX_DICT_SIZES
    #     d = dict()
    #     if is_cluster1 and is_cluster2:
    #         d[self.data[_IDX_POINTS]] = 1
    #     elif is_cluster1:
    #         d = group2[i].copy()
    #         k = group1[_IDX_POINTS]
    #         d[k] = d.get(k, 0) + 1
    #     elif is_cluster2:
    #         d = group1[i].copy()
    #         k = group2[_IDX_POINTS]
    #         d[k] = d.get(k, 0) + 1
    #     else:
    #         d = group1[i].copy()
    #         for k in group2[i]:
    #             d[k] = d.get(k, 0) + group2[i][k]
    #     self.data[i] = d




# cython: language_level=3
# cython: boundscheck=False
# cython: nonecheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True

# Labels the nodes using buffers from previous DRUHG's results.
# Also provides tools for label manipulations, such as:
# * Treats small clusters as outliers
# * (on-demand) Breaks big clusters
# * (on-demand) Glues outliers to the nearest clusters
#
# Author: Pavel Artamonov
# License: 3-clause BSD

import numpy as np
cimport numpy as np

from ._druhg_unionfind import UnionFind
from ._druhg_unionfind cimport UnionFind

from ._druhg_group cimport set_precision
from ._druhg_group import Group
from ._druhg_group cimport Group

def allocate_buffer_labels(np.intp_t size):
    return np.empty(size, dtype=np.intp)

cdef class Clusterizer(object):
    cdef UnionFind _U
    cdef np.ndarray _values_arr
    cdef np.ndarray _data_arr # for motion only
    cdef np.ndarray group_arr
    cdef np.ndarray ret_sizes
    cdef np.ndarray ret_clusters

    def __init__(self, np.ndarray _uf_arr, int _size, np.ndarray _values_arr, np.ndarray _data_arr,
                np.ndarray buf_ret_clusters,
                np.ndarray buf_ret_sizes,
                np.ndarray buf_group_arr):
        """ Uses the results of DRUHG MST-tree algorithm(unionfind structure and values).
            Emerge clusters and sizes arrays for later labeling.

            Parameters
            ----------
            _uf_arr : ndarray
                Unionfind structure from first phase.

            _size : int
                Amount of nodes.

            _values_arr : ndarray
                Edge values two for each edge.

            _data_arr : ndarray, nullable
                Points coordinates (motion only).

            buf_ret_clusters : ndarray
                Buffer related to 2nd half of UF. If positive then parent clusters it.

            buf_ret_sizes : ndarray
                Buffer related to 2nd half of UF. Node size (points, not edges).

            buf_group_arr : ndarray
                Buffer related to 2nd half of UF.

        """

        if _size is None or _size == 0:
            _size = int((len(_uf_arr) + 1) / 2)
        self._U = UnionFind(_size, _uf_arr)

        self._data_arr = _data_arr
        self._values_arr = _values_arr

        # TODO: check the allocations and the size
        # if ret_labels is not None and len(ret_labels) < self._U.p_size:
        #     print('ERROR: labels buffer is too small', len(ret_labels), self._U.p_size)
        #     return

        self.group_arr = buf_group_arr
        self.ret_sizes = buf_ret_sizes
        self.ret_clusters = buf_ret_clusters


    cpdef emerge(self, precision=0.0000001, run_motion = False):
        cdef:
            np.intp_t p, i, cluster_size, loop_size, \
                offset = self._U.get_offset()
            np.double_t v, limit = 0.

        set_precision(precision)

        # if self._data_arr is None and run_motion:
        #     print('ERROR: data values are missing')
        #     return

        self.group_arr[:self._U.p_size].fill(0)
        self.ret_clusters[:self._U.p_size].fill(0)
        self.ret_sizes[:self._U.p_size].fill(0)

        # helpers
        x_group = Group(np.zeros_like(self.group_arr[:1])[0])
        y_group = Group(np.zeros_like(self.group_arr[:1])[0])
        p_group = Group(np.zeros_like(self.group_arr[:1])[0])
        outlier_group = Group(np.zeros_like(self.group_arr[:1])[0])
        outlier_group_cluster = 0
        outlier_group.cook_outlier(0)

        loop_size1, loop_size2 = self._U.p_size, self._U.p_size * 2

        for u in range(loop_size1):
            p = self._U.parent[u]
            if p == 0:  # in case a point doesnt have any connection
                continue
            assert p >= self._U.p_size
            p = p - offset
            assert 0 <= p <= self._U.p_size
            v = self._values_arr[p]
            # r = self.ranks_arr[has_ranks * p]

            # first ever node connection
            p_group.assume_data(self.group_arr[p], self.ret_sizes[p], self.ret_clusters[p])
            p_group.child(u, p_group.points() == 0)

            if run_motion:
                p_group.mtn_add_1_coords(self._data_arr[u], v)

            p_group.add_1_autocluster(v)
            self.ret_sizes[p] = p_group.points()
            self.ret_clusters[p] = p_group.uniq_edges()

        for u in range(loop_size1 + 1, loop_size2):
            i = u - offset
            assert 0 <= i <= self._U.p_size

            if self.ret_sizes[i] == 0:
                break
            x_group.assume_data(self.group_arr[i], self.ret_sizes[i], self.ret_clusters[i])

            p = self._U.parent[u]
            if p == 0:
                continue
            p = p - offset
            assert 0 <= p <= self._U.p_size

            p_group.assume_data(self.group_arr[p], self.ret_sizes[p], self.ret_clusters[p])

            has_child, is_outlier, j = p_group.child(i, False)
            if not has_child:  # save to process later as pair
                continue

            v = self._values_arr[p]

            if is_outlier:
                y_group.assume_data(outlier_group.data, 1, 0)  # плохо сделано, надо отдельный метод
                y_is_cluster = True
            else:
                y_group.assume_data(self.group_arr[j],
                                    self.ret_sizes[j],
                                    self.ret_clusters[j])  # можем таргетить дефолта? Когда координаты не нужны - 100%
                y_is_cluster = False

            common_coef = 1. / (x_group.uniq_edges() + y_group.uniq_edges() - 1.)
            # print('common_coef', common_coef)
            x_is_cluster = x_group.will_cluster(v, common_coef * y_group.points())
            if not is_outlier:
                y_is_cluster = y_group.will_cluster(v, common_coef * x_group.points())

            p_group.aggregate(v, x_is_cluster, x_group, y_is_cluster, y_group)

            self.ret_sizes[p] = p_group.points()
            self.ret_clusters[p] = p_group.uniq_edges()

            assert x_group.uniq_edges() <= 0
            # non-clusters are negative
            self.ret_clusters[i] = -x_group.uniq_edges() if x_is_cluster else x_group.uniq_edges()
            if not is_outlier:
                assert y_group.uniq_edges() <= 0
                self.ret_clusters[j] = -y_group.uniq_edges() if y_is_cluster else y_group.uniq_edges()

            # if run_motion:
            #     # outlier coords already in p_group
            #     p_group.mtn_aggregate(v, x_is_cluster, x_group.data, y_is_cluster, y_group.data)
            #     x_group.mtn_mark_cluster(x_is_cluster)
            #     y_group.mtn_mark_cluster(y_is_cluster)

        if run_motion:
            return self.group_arr
        return self.ret_clusters, self.ret_sizes

    cdef void _fixem(self, np.ndarray edges_arr, np.intp_t num_edges, np.ndarray result):
        cdef:
            np.intp_t p, a, b, dontstop
            set new_results, links
            list new_path, restart

        new_results = set()
        new_path = []
        restart = []
        for p in range(0, num_edges):
            a, b = edges_arr[2 * p], edges_arr[2 * p + 1]
            if result[a] < 0 and result[b] < 0:
                new_results.update([a, b])
                new_path.append((a, b))
                continue
            elif result[b] < 0:
                a, b = b, a
            elif result[a] >= 0:
                continue
            res = result[b]
            result[a] = res
            if a in new_results:
                links = set([a])
                dontstop = 1
                while dontstop:
                    dontstop = 0
                    for path in list(new_path):
                        a, b = path
                        if a in links or b in links:
                            result[a] = result[b] = res
                            links.update([a, b])
                            new_path.remove(path)
                            dontstop = 1

        return

    cdef _mark_labels(self, ret_labels,
                     list exclude = None,
                      np.intp_t limitL = 0, np.intp_t limitH = 0,
                     ):
        cdef np.intp_t i, p, pp, label, offset

        offset = self._U.get_offset()

        i = self._U.p_size
        while i:
            i -= 1
            p = self._U.parent[i]
            label = -1
            while p != 0:
                pp = p - offset
                cluster_size = self.ret_sizes[pp]
                if cluster_size > limitH:
                    break
                if self.ret_clusters[pp] > 0 and cluster_size >= limitL and pp not in exclude:
                    label = pp
                p = self._U.parent[p]
            ret_labels[i] = label
        return ret_labels

    cpdef np.ndarray label(self, np.ndarray ret_labels,
                                list exclude=None, size_range=None,
                                np.intp_t fix_outliers=0, edgepairs_arr=None,
                                precision=0.0000001):
        """Returns cluster labels and clusters densities.
    
        Marks data-points with corresponding parent index of a cluster.
        Exclude list breaks passed clusters by their parent index.
        `size_range` breaks clusters outside it's range.
        Outliers-noise marked by -1.
    
        Parameters
        ----------
    
        ret_labels : ndarray
            The result. -1 are outliers.    
    
        exclude : list
            Clusters with parent-index from this list will not be formed. 
    
        size_range : list, optional (default=[1,size])
            Clusters that are smaller or over than the range treated as noise. 
            Pass None to find True outliers.
    
        fix_outliers: bool, optional (default=False)
            All outliers will be assigned to the nearest cluster. Need to pass mst(edgepairs).
    
        edgepairs_arr: array, optional (default=None)+
            Used with fix_outliers.
            
        precision: double, optional
            Relevant for small distances
    
        Returns
        -------
    
        labels : array [size]
           An array of cluster labels, one per data-point. Unclustered points get
           the label -1.
    
        metalabels : dictionary, on-demand
            A dictionary: keys - labels, values - tuples (distance, rank).   
    
        """
        cdef:
            int i

        if size_range is None:
            limitL, limitH = 0, 0
        else:
            limitL, limitH = size_range[0], size_range[1]
        if limitL < 0 or limitL > self._U.p_size:
            print ('label: size_range[0] is ignored. Cannot use '+str(limitL))
            limitL = 0
        if limitL < 1:
            limitL = int(limitL * self._U.p_size)

        if limitH <= 0 or limitH > self._U.p_size + 1:
            print ('label: size_range[1] is ignored. Cannot use ' + str(limitH))
            limitH = self._U.p_size
        if limitH <= 1:
            limitH = int(limitH * self._U.p_size + 1)

        if not exclude:
            exclude = []

        if ret_labels is not None and len(ret_labels) < self._U.p_size:
            print('ERROR: labels buffer is too small', len(ret_labels), self._U.p_size)
            return

        ret_labels = self._mark_labels(ret_labels,
                                 exclude, limitL, limitH)

        if fix_outliers == 1 and len(np.unique(ret_labels)) > 1:
            if edgepairs_arr is not None:
                self._fixem(edgepairs_arr, self._U.p_size - 1, ret_labels)
            else:
                print('To fix_outliers pass edgepairs', edgepairs_arr)

        return ret_labels


cdef np.ndarray pretty(np.ndarray labels_arr):
    """ Relabels to pretty positive integers. 
    """
    cdef np.intp_t i, p, label, max_label
    cdef np.ndarray[np.intp_t, ndim=1] result_arr
    cdef dict converter
    cdef np.intp_t* result

    result_arr = -1*np.ones(len(labels_arr), dtype=np.intp)
    result = (<np.intp_t *> result_arr.data)

    converter = {-1: -1}
    max_label = 0
    i = len(labels_arr)
    while i:
        i -= 1
        p = labels_arr[i]
        if p in converter:
            label = converter[p]
        else:
            label = max_label
            converter[p] = max_label
            max_label += 1
        result[i] = label

    return result_arr



# -*- coding: utf-8 -*-
# Author: Pavel Artamonov
# License: 3-clause BSD


import numpy as np
import collections
from math import exp, log

# from scipy.cluster.hierarchy import dendrogram
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from warnings import warn

class UF(object): # shadows _druhg_unionfind
    def __init__(self, parents_arr, size):
        self.parent = parents_arr
        self.p_size = size

    def get_offset(self):
        return self.p_size + 1


class ClusterTree(object):
    def __init__(self, uf_arr, data_arr, values_arr=None, sizes_arr=None, clusters_arr=None, mst_pairs=None, interactive=False):
        self._U = UF(uf_arr, len(data_arr))
        self._raw_data = data_arr
        self._values_arr = values_arr
        self._sizes_arr = sizes_arr
        self._clusters_arr = clusters_arr

        self._static_labels = None

        self._mst_pairs = mst_pairs # TODO: rebuild it if null?
        self._sum_coords = None

        self.clusters_pallete_ = np.zeros(len(self._raw_data), (np.double, 4))
        self.node_colors_ = np.zeros(len(self._raw_data), (np.double, 4))
        self.scat_ = None
        self.quiver_ = None
        self.quiver_colors_ = None
        self.annotation_ = None
        self.outlier_color_ = None
        self._timer_text = None

    def decrease_dimensions(self):
        if self._raw_data.shape[1] > 2:
            # Get a 2D projection; if we have a lot of dimensions use PCA first
            if self._raw_data.shape[1] > 32:
                # Use PCA to get down to 32 dimension
                data_for_projection = PCA(n_components=32).fit_transform(self._raw_data)
            else:
                data_for_projection = self._raw_data

            projection = TSNE().fit_transform(data_for_projection)
        elif self._raw_data.shape[1] == 2:
            projection = self._raw_data.copy()
        else:
            # one dimensional. We need to add dimension
            projection = self._raw_data.copy()
            projection = np.array([e for e in enumerate(projection)], np.int)

        return projection

    def get_cluster(self, e, top_dis, range_size):
        ret_index = -1
        while self._U.parent[e] != 0:
            p = self._U.parent[e]
            pc = p - self._U.get_offset()
            if self._sizes_arr[pc] > range_size[1]:
                break
            if top_dis < self._values_arr[pc]:
                break
            if self._clusters_arr[pc] > 0:  # it is a cluster
                if range_size[0] <= self._sizes_arr[pc]:
                    ret_index = pc
            e = p
        return ret_index

    def _avg_color(self, colora, colorb):
        color = (colora + colorb) / 2.
        return color

    def _plot_edges(self, ax, pos, node_colors, top_dis):
        try:
            from matplotlib import collections as mc
            from matplotlib.pyplot import Arrow
        except ImportError:
            raise ImportError('You must install the matplotlib library to plot the minimum spanning tree.')

        num_edges = len(self._mst_pairs) // 2
        start, end = pos[self._mst_pairs[num_edges//2]], pos[self._mst_pairs[num_edges//2 + 1]]
        if self.quiver_ is None:
            line_width = min(1., ((start[0] - end[0]) ** 2 + (start[1] - end[1]) ** 2) ** 0.5) / 2.  # медианный размер
            x, y, u, v = [], [], [], []
            for i in range(0, num_edges):
                a, b = self._mst_pairs[2*i], self._mst_pairs[2*i+1]
                start, end = pos[a], pos[b]
                x.append(start[0])
                y.append(start[1])
                u.append(end[0] - start[0])
                v.append(end[1] - start[1])
            self.quiver_ = ax.quiver(x, y, u, v, angles='xy', scale_units='xy', scale = 1)

        if self.quiver_colors_ is None:
            self.quiver_colors_ = np.zeros(len(self._raw_data), (np.double, 4))
        for i in range(0, num_edges):
            a, b = self._mst_pairs[2*i], self._mst_pairs[2*i+1]
            color = self._avg_color(node_colors[a], node_colors[b])
            self.quiver_colors_[i][0] = color[0]
            self.quiver_colors_[i][1] = color[1]
            self.quiver_colors_[i][2] = color[2]
            self.quiver_colors_[i][3] = color[3] * 0.75
            if self._values_arr[i] >= top_dis:
                self.quiver_colors_[i][3] = 0.
            self.quiver_.set_color(self.quiver_colors_)

    def convert_labels_to_colors(self, palette, base_node_alpha):
        different_colors = len(palette)
        num_points = len(self._static_labels)
        for i in range(0, num_points):
            lbl = self._static_labels[i]
            if lbl < 0:
                self.clusters_pallete_[i] = self.outlier_color_
            else:
                new_col = palette[lbl % different_colors]
                self.clusters_pallete_[i][0] = new_col[0]
                self.clusters_pallete_[i][1] = new_col[1]
                self.clusters_pallete_[i][2] = new_col[2]
                self.clusters_pallete_[i][3] = base_node_alpha

    def bg_colors_and_pallete(self, palette, base_node_alpha):
        different_colors = len(palette)
        rev_color = 0
        num_points = len(self._raw_data)
        size_uf = len(self._U.parent)
        slider_sizes_bg = np.zeros(num_points + 1)
        for i in range(num_points+1, size_uf):
            if self._U.parent[i] == 0:
                continue
            cl = i - self._U.get_offset()
            pc = self._U.parent[i] - self._U.get_offset()
            if self._clusters_arr[pc] > 0:
                slider_sizes_bg[self._sizes_arr[pc]] += 1

            pc = self._U.parent[i] - self._U.get_offset()
            col = self.clusters_pallete_[pc]
            new_col = palette[rev_color]
            # making sure that parent has different color
            if col[0] == new_col[0] and col[1] == new_col[1] and col[2] == new_col[2]:
                new_col = palette[rev_color]

            self.clusters_pallete_[pc][0] = new_col[0]
            self.clusters_pallete_[pc][1] = new_col[1]
            self.clusters_pallete_[pc][2] = new_col[2]
            self.clusters_pallete_[pc][3] = base_node_alpha

            rev_color += 1
            if rev_color >= different_colors:
                rev_color = 0

        for i in range(0, num_points+1):
            if slider_sizes_bg[i] == 0: # making more visible on the axis
                slider_sizes_bg[i] = np.nan
        return slider_sizes_bg

    def plot(self, static_labels=None, axis=None, interactive=True,
             node_size=40, node_color=None,
             node_alpha=0.8, edge_alpha=0.15, edge_linewidth=8,
             core_color='purple'):
        """Plot the cluster tree with slider controls.

        Parameters
        ----------
        static_labels : array, optional
                If passed - no slider widgets

        axis : matplotlib axis, optional
                The axis to render the plot to

        node_size : int, optional (default 40)
                The size of nodes in the plot.

        node_color : matplotlib color spec, optional
                By default draws colors according to labels
                where alpha regulated by cluster size.

        node_alpha : float, optional (default 0.8)
                The alpha value (between 0 and 1) to render nodes with.

        edge_alpha : float, optional (default 0.4)
                The alpha value (between 0 and 1) to render nodes with.

        edge_linewidth : float, optional (default 2)
                The linewidth to use for rendering edges.

        core_color : matplotlib color spec, optional (default 'purple')
                Plots colors at the node centers.
                Can be omitted by passing None.

        Returns
        -------

        axis : matplotlib axis
                The axis used the render the plot.
        """
        try:
            import matplotlib.pyplot as plt
            from matplotlib.widgets import Slider, Button, RangeSlider
        except ImportError:
            raise ImportError('You must install the matplotlib library to plot cluster tree.')

        try:
            import datetime
        except ImportError:
            raise ImportError('You must install the datetime library to plot cluster tree.')

        try:
            import seaborn as sns
        except ImportError:
            raise ImportError('You must install the seaborn library to draw colored labels.')

        if self._raw_data.shape[0] > 32767:
            warn('Too many data points for safe rendering of a cluster tree!')
            return None

        self._static_labels = static_labels

        pos = self.decrease_dimensions()

        # рисуем полотно и выводим два слайдера
        # выводить статы по времени отрисовки
        #   если время превышает, то выводить кнопку Апплай и выводить только при её нажатии
        dis_slider, qty_slider, btn_apply = None, None, None
        fig = None

        different_colors = 10
        base_node_alpha = 0.8
        def restricted_labeling(top_dis, range_size):
            num_points = len(self._raw_data)
            for i in range(0, num_points):
                pc = self.get_cluster(i, top_dis, range_size)
                if pc > 0:
                    self.node_colors_[i] = self.clusters_pallete_[pc]
                    # num_clusters += 1
                else:
                    self.node_colors_[i] = self.outlier_color_
            return self.node_colors_

        def on_pick(event):
            annotation_visible = self.annotation_.get_visible()
            if event.inaxes == axmain:
                is_contained, annotation_index = self.scat_.contains(event)
                if is_contained:
                    point_loc = self.scat_.get_offsets()[annotation_index['ind'][0]]
                    self.annotation_.xy = point_loc
                    ind = annotation_index['ind'][0]
                    pc = self.get_cluster(ind, self._values_arr[int(dis_slider.val)]*1.0001, qty_slider.val)
                    ss = 1
                    cl = 0
                    dis = 0
                    if pc >= 0:
                        ss = self._sizes_arr[pc]
                        cl = self._clusters_arr[pc] + 1
                        dis = self._values_arr[pc]
                    text_label = 'label:' + str(pc) \
                                 + '\n dis: {:.4f}'.format(dis) \
                                 + '\n size: ' + str(ss) \
                                 + '\nparts: ' + str(cl) \
                                 + '\n(p: '+ str(ind)+')'
                    self.annotation_.set_text(text_label)
                    self.annotation_.set_visible(True)
                    fig.canvas.draw_idle()
                else:
                    if annotation_visible:
                        self.annotation_.set_visible(False)
                        fig.canvas.draw_idle()

        def update_plot(val):
            # axis.cla()
            now = datetime.datetime.now()

            if dis_slider is None:
                dis = np.inf
            else:
                dis = self._values_arr[int(dis_slider.val)]*1.0001

            if qty_slider is None:
                range_ = [0, np.inf]
            else:
                range_ = qty_slider.val

            if self._static_labels is None:
                cc = restricted_labeling(dis, range_)
            else:
                cc = self.clusters_pallete_

            if self._mst_pairs is not None:
                self._plot_edges(axmain, pos, cc, dis)  # edge_linewidth, edge_alpha, vary_line_width)

            if self.scat_ is None:
                self.scat_ = axmain.scatter(pos.T[0], pos.T[1], c=cc, s=node_size, alpha=node_alpha)
                axmain.set_axis_off()

                if fig is not None and self._static_labels is None:
                    self.annotation_ = axmain.annotate(
                        text='',
                        xy=(0,0),
                        xytext=(10,15),
                        textcoords='offset points',
                        bbox={'boxstyle': 'round'},
                        arrowprops={'arrowstyle': '->'}
                    )
                    self.annotation_.set_visible(False)
                    # fig.canvas.mpl_connect('motion_notify_event', motion_hover)
                    fig.canvas.mpl_connect('button_press_event', on_pick)

                if core_color is not None:
                    # adding (red)dots at the node centers
                    axmain.scatter(pos.T[0], pos.T[1], c=core_color, marker='.', s=node_size / 10)
            else:
                self.scat_.set_color(cc)

            td = datetime.datetime.now() - now
            if self._static_labels is None:
                if self._timer_text is None:
                    self._timer_text = axmain.text(0.05, 0.95, f'{td.total_seconds():.3f}' + " sec", transform=plt.gcf().transFigure,
                                                   verticalalignment='top', horizontalalignment='left')
                else:
                    self._timer_text.set_text(f'{td.total_seconds():.3f}' + " sec")
            if self.scat_ is not None and btn_apply is not None and  not axbtn.get_visible():
                if td.total_seconds() < 3:
                    axbtn.set_visible(False)
                else:
                    axbtn.set_visible(True)

        if axis is not None:
            axmain = axis
        elif self ._static_labels is not None:
            # fig = plt.figure()
            axmain = plt.gca()
            axmain.set_axis_off()
        else:
            fig, axs = plt.subplots(2, 2, width_ratios=[0.9, 0.1], height_ratios=[0.95, 0.05])
            axmain = axs[0, 0]

        self.outlier_color_ = axmain.get_facecolor()
        self.outlier_color_ = (1. - self.outlier_color_[0], 1. - self.outlier_color_[1], 1. - self.outlier_color_[2], 0.5)

        if self._static_labels is not None:
            self.convert_labels_to_colors(sns.color_palette('bright', different_colors+2), base_node_alpha)
        else:
            slider_sizes_bg = self.bg_colors_and_pallete(sns.color_palette('bright', different_colors+2), base_node_alpha)

        if axis is not None or self ._static_labels is not None:
            update_plot(None)
            # plt.show()
            return axmain

        # dynamic with sliders
        axvals = axs[1, 0]
        axqty = axs[0, 1]
        axbtn = axs[1, 1]
        num_points = len(self._raw_data)

        def update_qty_slider(val):
            qty_slider.poly.set_xy([[0, 0], [1, 0],
                                   [1, qty_slider.val[0]], [0, qty_slider.val[0]],
                                   [0, qty_slider.val[1]], [1, qty_slider.val[1]],
                                   [1, num_points], [0, num_points]])

            if val is not None and btn_apply is not None and not axbtn.get_visible():
                update_plot(val)
            fig.canvas.draw_idle()
            fig.canvas.flush_events()

        def update_dis_slider(val):
            dis = self._values_arr[int(dis_slider.val)]
            dis_slider.valtext.set_text("{:.4f}".format(dis))
            # dis_slider.poly.set_xy([[dis_slider.val, 0.], [dis_slider.val, 2.], [num_points, 2.], [num_points, 0.]])
            dis_slider.poly.set(xy=[dis_slider.val, 0.], height=2., width=(num_points - dis_slider.val + 1))

            if val is not None and btn_apply is not None and not axbtn.get_visible():
                update_plot(val)
            if btn_apply is None or not axbtn.get_visible():
                fig.canvas.draw_idle()
            fig.canvas.flush_events()

        def on_key_press(event):
            if event.key == 'left' and dis_slider.val > 0:
                dis_slider.set_val(dis_slider.val - 1)
            elif event.key == 'right' and dis_slider.val < dis_slider.valmax:
                dis_slider.set_val(dis_slider.val + 1)

            v1, v2 = qty_slider.val
            if event.key == 'up' and v1+1 < v2:
                qty_slider.set_val([v1+1, v2])
            elif event.key == 'down' and v2 > 0:
                qty_slider.set_val([v1 - 1, v2])
            elif event.key == 'shift+up' and v1 < dis_slider.valmax:
                qty_slider.set_val([v1, v2 + 1])
            elif event.key == 'shift+down' and v1+1 < v2:
                qty_slider.set_val([v1, v2-1])

        def _apply(event):
            axbtn.set_visible(False)
            update_plot(None)
            fig.canvas.draw_idle()
            fig.canvas.flush_events()

        axvals.plot(self._values_arr, scaley='log', color=self.outlier_color_)
        dis_slider = Slider(axvals, 'Values', valmin=0, valmax=num_points-2,
                            valstep=1.,
                            valinit=num_points-2,
                            color=(self.outlier_color_[0],self.outlier_color_[1],self.outlier_color_[2], 0.2),
                            track_color=(0.5, 0.5, 0.5, 0.05),
                            handle_style={"": "|", "size": 30},
                            )

        axqty.plot(slider_sizes_bg, range(0, len(slider_sizes_bg)), 'k_', scalex='log', color=self.outlier_color_)
        qty_slider = RangeSlider(axqty, "Qty", valmin=0, valmax=num_points,
                                 valstep=1., orientation="vertical",
                                 color=(self.outlier_color_[0],self.outlier_color_[1],self.outlier_color_[2],0.2),
                                 track_color=(0.5, 0.5, 0.5, 0.05),
                                 handle_style={"": "_", "size": 30},
        )

        btn_apply = Button(axbtn, 'Run', color='gray', hovercolor='green')
        btn_apply.on_clicked(_apply)
        axbtn.set_visible(False)

        dis_slider.on_changed(update_dis_slider)
        qty_slider.on_changed(update_qty_slider)
        cid = fig.canvas.mpl_connect('key_press_event', on_key_press)

        # init
        update_dis_slider(None)
        update_qty_slider(None)
        update_plot(None)

        plt.show()

        return axmain



"""
Tests for DRUHG clustering algorithm
"""
# pytest -k "test_name"
import pickle
import numpy as np
import pandas as pd
from scipy.spatial import distance
from scipy import sparse
from scipy import stats
import pytest  #delete __init__.py or it wont work
import time


from druhg import (DRUHG,
                   druhg)

from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# from sklearn.cluster.tests.common import generate_clustered_data
import sklearn.datasets as datasets
from sklearn.datasets import make_blobs
from sklearn.utils import shuffle
from sklearn.preprocessing import StandardScaler
from scipy.stats import mode
from sklearn.metrics import adjusted_rand_score

from tempfile import mkdtemp
from functools import wraps

import warnings

moons, _ = datasets.make_moons(n_samples=50, noise=0.05)
blobs, _ = datasets.make_blobs(n_samples=50, centers=[(-0.75, 2.25), (1.0, 2.0)], cluster_std=0.25)
X = np.vstack([moons, blobs])

_plot_graph = 1
_test_extra_visualisation = 1
_not_fail_all = True

def test_minitest(filename=None):
    if filename is None:
        filename = test_minitest.__name__
    XX = np.array([[ 9.82287700e-01, -4.91767541e-01],
       [ 7.79582720e-01,  6.08296786e-01],
       [-8.88193604e-01,  4.43296052e-01],
       [ 7.78691236e-02,  1.05579835e-01],
       [ 7.05833295e-01, -4.49687952e-01],
       [ 9.08054027e-01,  3.82957634e-01],
       [-8.44357575e-01,  4.76983140e-01],
       [-6.96863884e-01,  7.50031808e-01],
       [ 4.39031497e-01,  8.96900018e-01],
       [ 2.06198639e+00,  4.95933426e-01],
       [ 2.00697002e+00,  2.51940542e-01],
       [ 8.86671509e-01,  5.33899191e-01],
       [ 1.01472887e+00,  1.77887023e-01],
       [-9.99361908e-01,  8.14805761e-02],
       [ 1.66364569e+00, -2.72722628e-01],
       [ 1.45444706e-01, -3.60310809e-04],
       [ 2.47847446e-02,  2.47518860e-01],
       [ 5.33608226e-01, -4.16318654e-01],
       [-1.02367914e+00,  2.03681301e-01],
       [ 1.72252995e+00, -1.73532360e-01],
       [ 1.19371766e-01,  9.66962174e-01],
       [-9.96385260e-01,  1.51193473e-03],
       [ 1.97170679e+00,  1.10284595e-01],
       [-4.04580401e-03,  3.53573375e-01],
       [ 6.27338990e-01, -4.64420966e-01],
       [ 1.71681386e-01, -1.17028174e-01],
       [ 1.26919122e+00, -4.46221578e-01],
       [ 6.14585010e-02,  1.00584911e+00],
       [ 1.09577989e+00, -5.01458976e-01],
       [ 3.96411149e-01, -2.98249260e-01],
       [ 1.88598729e+00,  7.33595903e-03],
       [ 1.96974012e+00,  3.38327527e-01],
       [-5.08398106e-01,  8.60840352e-01],
       [-4.00736795e-01,  8.89316463e-01],
       [ 1.77070595e+00, -1.36976988e-01],
       [ 1.49886531e+00, -3.96515486e-01],
       [ 7.22042880e-01,  7.04074310e-01],
       [ 8.49944965e-01, -5.34124448e-01],
       [-9.09817816e-02,  1.03513561e+00],
       [ 5.14565924e-01,  7.87485818e-01],
       [ 4.02652800e-01,  9.37234351e-01],
       [-2.91295702e-01,  9.02204623e-01],
       [ 2.87744167e-01, -2.31116381e-01],
       [-3.31091977e-02,  4.88892172e-01],
       [ 2.41871950e-01,  9.92840358e-01],
       [ 1.06415165e+00, -1.65735096e-02],
       [ 1.37323752e+00, -4.65626874e-01],
       [-6.29552626e-01,  7.87828922e-01],
       [-8.27252851e-01,  5.81983413e-01],
       [ 1.00280073e+00,  2.68063630e-01],
       [-7.83496058e-01,  2.18322479e+00],
       [-9.92609950e-01,  2.02124136e+00],
       [-9.81354504e-01,  2.29182851e+00],
       [-4.98266423e-01,  1.78721081e+00],
       [-1.08969024e+00,  2.32889766e+00],
       [-1.08081235e+00,  1.98253144e+00],
       [-1.09687394e+00,  1.95677593e+00],
       [-1.00962708e+00,  1.98160320e+00],
       [-5.23444712e-01,  2.05973030e+00],
       [-5.27378118e-01,  2.01577480e+00],
       [-1.01076274e+00,  2.26189991e+00],
       [-1.07307453e+00,  1.59463042e+00],
       [-3.71408468e-02,  2.36246794e+00],
       [-4.78906511e-01,  2.09108831e+00],
       [-1.36496960e+00,  1.75639407e+00],
       [-9.63722898e-01,  1.77177389e+00],
       [-8.57835037e-01,  2.85261536e+00],
       [-9.01706144e-01,  1.96899908e+00],
       [-7.83591885e-01,  1.97077401e+00],
       [-7.53906201e-01,  2.01930595e+00],
       [-1.06416195e+00,  2.09696002e+00],
       [-1.06425933e+00,  1.92864353e+00],
       [-5.59921254e-01,  1.92873234e+00],
       [-2.93844091e-01,  2.13445936e+00],
       [-1.08927720e+00,  2.22384623e+00],
       [-5.47674986e-01,  2.56744321e+00],
       [-8.32079069e-01,  1.91501237e+00],
       [-8.64079731e-01,  2.02405299e+00],
       [-9.43370303e-02,  1.53264890e+00],
       [-8.46705805e-01,  2.08845880e+00],
       [-7.64153884e-01,  1.95806975e+00],
       [-7.97244563e-01,  1.93409856e+00],
       [-6.11964372e-01,  1.83346591e+00],
       [-7.97508190e-01,  2.18510057e+00],
       [-5.06448969e-01,  1.99409096e+00],
       [-6.53243323e-01,  2.52892190e+00],
       [-8.73821576e-01,  1.80817140e+00],
       [-7.03121501e-01,  1.95331014e+00],
       [-9.83658309e-01,  2.33860746e+00],
       [-1.01683507e+00,  1.94204773e+00],
       [-4.87343036e-01,  1.56494127e+00],
       [-7.79832602e-01,  2.07966975e+00],
       [-4.46877995e-01,  1.45948349e+00],
       [-3.51769879e-01,  1.59343158e+00],
       [-1.34234725e+00,  2.31765803e+00],
       [-9.56682748e-01,  1.53054479e+00],
       [-6.10630098e-01,  1.87378092e+00],
       [-7.42290479e-01,  1.84433031e+00],
       [-8.39334432e-01,  1.81906771e+00],
       [-9.27731599e-01,  2.29639091e+00],
       [ 1.42071742e+00,  3.30486324e+00],
       [ 1.55862485e+00,  1.30863477e+00],
       [ 1.24972581e+00,  2.42336599e+00],
       [ 5.33685073e-01,  2.59954396e+00],
       [ 7.68000568e-01,  2.08111755e+00],
       [ 2.35295779e+00,  3.68309710e+00],
       [ 1.54597101e+00,  1.84281151e+00],
       [ 8.12131013e-01,  3.29007476e+00],
       [ 7.94516457e-01,  2.41783502e+00],
       [ 7.98220091e-01,  9.82325962e-01],
       [ 4.41618515e-01,  2.86576268e+00],
       [ 2.28132907e+00,  2.62941621e+00],
       [ 1.30744465e+00,  9.21494187e-01],
       [ 2.89402898e+00,  3.07162590e+00],
       [ 2.59828872e+00,  3.48249198e+00],
       [ 2.39156834e-01,  1.13332844e+00],
       [ 1.28134650e+00,  1.69771109e+00],
       [ 2.47539905e+00,  3.84649616e+00],
       [ 8.47050787e-01,  2.07370998e+00],
       [ 1.70839331e+00,  2.03061588e+00],
       [ 1.86776491e+00,  2.64855805e+00],
       [ 1.82506092e+00,  2.65118340e+00],
       [ 7.38734334e-01,  2.46762804e+00],
       [ 2.05216302e+00,  3.09939872e-01],
       [ 2.08265157e+00,  2.68029197e+00],
       [ 5.48953797e-01,  1.63082304e+00],
       [ 6.66897628e-01,  2.19248368e+00],
       [ 7.21620016e-02,  1.92199689e+00],
       [ 9.35013031e-01,  1.14139910e+00],
       [ 2.10701578e+00,  1.56207334e+00],
       [ 1.32685731e+00,  3.02292533e+00],
       [-3.12399518e-01,  4.03870510e+00],
       [ 2.45342595e+00,  3.36391798e+00],
       [ 7.40975862e-01,  1.28206197e+00],
       [ 7.23212885e-01,  1.25423311e+00],
       [ 3.20586402e+00,  2.00176070e+00],
       [ 5.44831804e-01,  8.42644367e-01],
       [ 4.91531506e-01,  8.54574838e-01],
       [ 1.09246967e+00,  2.28977328e+00],
       [ 1.34148473e+00,  3.59316431e+00],
       [ 1.47390241e+00,  2.68453871e+00],
       [ 2.04938161e+00,  3.00980325e-01],
       [ 4.74061748e-01,  1.82517865e+00],
       [ 1.38255750e+00,  1.49441056e+00],
       [-1.81396229e-01,  1.67608102e+00],
       [ 1.61734753e+00,  1.62344100e+00],
       [ 2.27547713e+00,  3.02798895e+00],
       [ 1.26536052e+00,  2.74417246e+00],
       [ 2.22312311e+00,  1.59609013e+00],
       [ 7.16857960e-01,  1.30979190e+00],
       [ 1.69435113e+00,  2.58428292e+00],
       [ 2.56004668e+00,  1.90043678e+00],
       [ 4.00593052e-01,  2.07202927e+00],
       [ 1.09514002e+00,  8.32401990e-01],
       [ 3.44335366e-01,  2.12210586e+00],
       [ 1.66737772e+00,  2.58321428e+00],
       [ 6.30088239e-01,  3.34926791e+00],
       [ 1.71452800e+00,  4.05206484e+00],
       [ 1.56123347e+00,  2.73666162e+00],
       [ 2.25416915e+00,  3.33381046e+00],
       [ 1.92722848e+00,  2.84539606e+00],
       [ 1.94938390e+00,  2.02922499e+00],
       [ 1.33211707e+00,  7.77595470e-01],
       [ 1.44282596e+00,  3.03895516e+00],
       [ 2.53818450e+00,  2.09297521e+00],
       [ 7.06781846e-01,  2.88379784e+00],
       [ 8.75255708e-01,  2.19440080e+00],
       [ 1.03602323e+00,  1.95794841e+00],
       [ 2.05569358e+00,  2.81113483e+00],
       [ 1.29284658e+00,  3.66966168e+00]])
    dr = DRUHG(max_ranking=50, verbose=False,
               do_edges=True, size_range=[1, 1])
               # do_edges=True, limitL=25, limitH=0.3, fix_outliers=1)

    dr.fit(XX)
    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'1'+'.png')
        # dr.plot()

    # if _plot_graph and _test_extra_visualisation:
    #     plt.close('all')
    #     dr.single_linkage_.plot(dr.labels_)
    #
    #     plt.savefig(filename +'plot'+ '.png')
    assert (False)

def test_plotcluster(filename=None):
    if filename is None:
        filename = test_plotcluster.__name__
    # XX = datasets.load_iris()['data']
    XX = np.array(pd.read_csv('druhg\\tests\\chameleon.csv', sep='\t', header=None))
    dr = DRUHG(max_ranking=50, verbose=False, do_edges=True)
    dr.fit(XX)
    if _test_extra_visualisation:
        plt.close('all')
        # plt.style.use('dark_background')
        # dr.plot(dr.labels_)
        dr.plot()
        plt.savefig(filename+'1'+'.png')
    assert (False)

def test_iris(filename=None):
    if filename is None:
        filename = test_iris.__name__
    iris = datasets.load_iris()
    XX = iris['data']
    # print(XX, type(XX))
    dr = DRUHG(max_ranking=50, verbose=False)
    dr.fit(XX)
    labels = dr.labels_
    if _plot_graph:
        plt.close('all')
        dr.plot(labels)
        plt.savefig(filename+'1'+'.png')

    ari = adjusted_rand_score(iris['target'], labels)
    print('iris ari', ari)
    # assert (ari >= 0.50)
    # breaking biggest cluster
    labels = dr.relabel(limitL=0, limitH=int(len(XX)/2), fix_outliers=True)

    if _plot_graph:
        plt.close('all')
        dr.plot(labels)
        plt.savefig(filename + '2' + '.png')

    ari = adjusted_rand_score(iris['target'], labels)
    print('iris ari', ari)
    assert (ari >= 0.85)
    assert _not_fail_all

def test_plot_mst():
    iris = datasets.load_iris()
    XX = iris['data']
    dr = DRUHG(max_ranking=50, do_edges=True)
    dr.fit(XX)
    dr.plot(dr.labels_)
    assert _not_fail_all

def test_plot_dendrogram(filename=None):
    if not _test_extra_visualisation:
        return
    if filename is None:
        filename = test_plot_dendrogram.__name__
    iris = datasets.load_iris()
    XX = iris['data']
    dr = DRUHG(max_ranking=50, limitH=int(len(XX)/2), fix_outliers=1)  #, limitL=0, limitH=int(len(XX)/2), fix_outliers=1)
    dr.fit(XX)
    if _plot_graph:
        plt.close('all')
        dr.single_linkage_.plot(dr.labels_)
        plt.savefig(filename+'.png')
    assert _not_fail_all

def test_plot_one_dimension():
    iris = datasets.load_iris()
    XX = iris['data']
    XX = XX.reshape(XX.points, 1)
    XX = np.array(sorted(XX))
    dr = DRUHG(max_ranking=50)
    dr.fit(XX)
    dr.plot(dr.labels_)
    assert _not_fail_all

def test_2and3():
    cons = 10.
    XX = [[0.,0.],[1.,1.],[cons+3.,2.],[cons+4.,1.],[cons+5.,2.]]
    XX = np.array(XX)
    dr = DRUHG(algorithm='slow', max_ranking=200, limitL = 1, limitH = 1000, verbose=False, do_edges=True)
    dr.fit(XX)
    # two clusters
    # assert (len(dr.parents_) == 2)
    print(dr.mst_)
    print(dr.mst_[6]*dr.mst_[7])

    labels = dr.labels_
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    n_clusters = len(set(labels)) - int(-1 in labels)
    print('n_clusters', n_clusters)

    # proper connection between two groups
    # assert (dr.mst_[6]*dr.mst_[7] == 2) # this is not working anymore

    assert (labels[0]==labels[1])
    assert (not all(x == labels[0] for x in labels))
    assert (labels[2] == labels[3] == labels[4])
    assert (labels[0] != labels[2])
    assert (n_clusters == 2)
    assert _not_fail_all

def test_scaled_flatrangle(filename=None):
    if filename is None:
        filename = test_scaled_flatrangle.__name__
    for i in range(-5, 4):
        test_line(size=3, scale=10 ** i, filename=filename, all_clusters=True)
    assert _not_fail_all

def test_rightangle(scale = 1., filename = None, all_clusters = True):
    if filename is None:
        filename = test_rightangle.__name__
    XX = [[0., 0.], [scale, 0], [0, scale]]
    XX = np.array(XX)
    dr = DRUHG(algorithm='slow', max_ranking=200, limitL=1, limitH=1000, verbose=False, do_edges=True)
    dr.fit(XX)
    # two clusters
    # assert (len(dr.parents_) == 2)

    labels = dr.labels_
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    n_clusters = len(set(labels)) - int(-1 in labels)
    print('n_clusters', n_clusters)

    assert (all(x == labels[0] for x in labels))
    assert _not_fail_all

def test_scaled_rightangle(scale = 1., filename = None, all_clusters = True):
    if filename is None:
        filename = test_scaled_rightangle.__name__
    for i in range(-5, 4):
        test_rightangle(scale=10 ** i, filename=filename, all_clusters=True)
    assert _not_fail_all

def test_equilateral(scale = 1., filename = None, all_clusters = True):
    if filename is None:
        filename = test_equilateral.__name__
    XX = [[scale, 0., 0.], [0., scale, 0], [0., 0., scale]]
    XX = np.array(XX)
    dr = DRUHG(algorithm='slow', max_ranking=200, limitL=1, limitH=1000, verbose=False, do_edges=True)
    dr.fit(XX)
    # two clusters
    # assert (len(dr.parents_) == 2)

    labels = dr.labels_
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    n_clusters = len(set(labels)) - int(-1 in labels)
    print('n_clusters', n_clusters)

    assert (all(x == labels[0] for x in labels))
    assert _not_fail_all

def test_scaled_equilateral(scale = 1., filename = None, all_clusters = True):
    if filename is None:
        filename = test_scaled_equilateral.__name__
    for i in range(-5, 4):
        test_equilateral(scale=10 ** i, filename=filename, all_clusters=True)
    assert _not_fail_all

def test_line(size=4, scale = 1., filename = None, all_clusters = False):
    if filename is None:
        filename = test_line.__name__
    XX = []
    for i in range(0,size):
        XX.append([0.,i*scale])
    XX = np.array(XX)

    dr = DRUHG(max_ranking=50, limitL=1, limitH=len(XX), verbose=False, do_edges=True)
    dr.fit(XX)
    # s = 2*len(XX) - 2
    # starts somewhere in the middle
    # and grows one by one
    # that's why there are no clusters
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    np.set_printoptions(precision=2)
    print('values', dr.values_)
    # assert (len(dr.parents_)==0)
    labels = dr.labels_
    if _plot_graph:
        plt.close('all')
        dr.plot(labels)
        plt.savefig(filename+'.png')

    if all_clusters:
        assert (all(x == labels[0] for x in labels))
        return

    assert (not all(x == labels[0] for x in labels))
    assert (labels[0] == labels[len(labels)-1])
    assert (labels[1] == labels[len(labels)-2])
    assert (labels[0] != labels[1])
    # assert _not_fail_all

def test_longline(filename=None):
    if filename is None:
        filename = test_longline.__name__
    test_line(size=100, filename=filename)
    assert _not_fail_all

def test_fiveline(filename=None):
    if filename is None:
        filename = test_fiveline.__name__
    test_line(size=5, filename=filename)
    assert _not_fail_all

def test_fourline(filename=None):
    if filename is None:
        filename = test_fourline.__name__
    test_line(size=4, filename=filename)
    assert _not_fail_all

# def test_hypersquare():
#     XX = []
#     size, scale = 6, 1.
#     for i1 in range(0, size):
#         for i2 in range(0, size):
#             for i3 in range(0, size):
#                 for i4 in range(0, size):
#                     for i5 in range(0, size):
#                         XX.append([i1*scale,i2*scale,i3*scale,i4*scale,i5*scale])
#     XX = np.array(XX)
#     dr = DRUHG(max_ranking=10, limitL=1, limitH=len(XX), verbose=False, do_edges=True)
#     dr.fit(XX)
#     s = 2*len(XX) - 2
#     print(dr.mst_)
#     print(dr.mst_[s-1], dr.mst_[s-2], XX[dr.mst_[s-1]], XX[dr.mst_[s-2]])
#     labels = dr.labels_
#     n_clusters = len(set(labels)) - int(-1 in labels)
#     print('n_clusters', n_clusters)
#     print(dr.mst_)
#     print(XX)
#     print(dr.labels_)
#     assert (n_clusters==1)
#     labels = dr.relabel(limitL=1)
#     n_clusters = len(set(labels)) - int(-1 in labels)
#     print('n_clusters', n_clusters)
#     assert (n_clusters == 5)
#     assert (0==1)

def test_square(showplot=True, size=10, scale=5., filename=None):
    if filename is None:
        filename = test_square.__name__
    XX = []
    for i in range(0, size):
        for j in range(0, size):
            XX.append([scale*i,scale*j])

    XX = np.array(XX)
    np.random.shuffle(XX)

    dr = DRUHG(max_ranking=2000, algorithm='slow',
               limitL=1, limitH=len(XX), verbose=False, do_edges=True)
    dr.fit(XX)

    if showplot and _plot_graph:
        plt.close('all')
        print('dr.labels_', dr.labels_)
        dr.plot(dr.labels_)
        # dr.plot(dr.labels_)
        plt.savefig(filename + 'plot' + '.png')

    s = 2 * len(XX) - 2
    # print(dr.mst_)
    # print(dr.mst_[s-1], dr.mst_[s-2], XX[dr.mst_[s-1]], XX[dr.mst_[s-2]])
    labels = dr.labels_
    n_clusters = len(set(labels)) - int(-1 in labels)
    print('n_clusters', n_clusters)
    print(dr.mst_)
    # print(XX)
    print(dr.labels_)
    # assert (n_clusters==1)
    # labels = dr.relabel(limitL=1, limitH=size*2)
    n_clusters = len(set(labels)) - int(-1 in labels)
    # print('n_clusters', n_clusters)
    # print('pairs', dr.mst_)
    print('labels', dr.labels_)

    un, cn = np.unique(labels, return_counts=True)
    for i in range(0, len(un)):
        print('square', un[i], cn[i] )

    labels = dr.labels_
    if showplot and _plot_graph:
        plt.close('all')
        dr.plot(labels)
        plt.savefig(filename+'.png')
    un, cn = np.unique(labels, return_counts=True)
    print('uniques', un, cn)

    sorteds = np.argsort(cn)
    print('hello')
    print('hello2', un[sorteds[0]], un[sorteds[1]])
    # assert (False)
    dr.relabel(limitL=1, limitH=size*2)
    if showplot and _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'2'+'.png')
    assert (n_clusters >= 5)
    assert (cn[sorteds[-1]] == (size-2)**2)
    assert (un[sorteds[0]] == -1 and cn[sorteds[0]] == 4)
    for i in range(1,5):
        assert (cn[sorteds[i]] == size - 2)
    assert _not_fail_all

def test_scaled_square(filename=None):
    if filename is None:
        filename = test_scaled_square.__name__
    for i in range(-4, 4):
        # test_square(scale=10 ** i, filename=filename + "i" + str(i) + "-")
        test_square(scale=10 ** i, filename=filename)
    assert _not_fail_all

def test_two_squares(filename=None):
    if filename is None:
        filename = test_two_squares.__name__
    XX = []
    size, scale = 6, 1
    for i in range(0, size):
        for j in range(0, size):
            XX.append([scale*i, scale*j])
            XX.append([2*size + scale*i, scale*j])
    XX = np.array(XX)
    dr = DRUHG(max_ranking=200, verbose=False, do_edges=True)
    dr.fit(XX)
    s = 2*len(XX) - 2
    print(dr.mst_)
    print(dr.mst_[s-1], dr.mst_[s-2], XX[dr.mst_[s-1]], XX[dr.mst_[s-2]])
    labels = dr.labels_
    n_clusters = len(set(labels)) - int(-1 in labels)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')
    assert (n_clusters==2)
    assert _not_fail_all


def test_particles(filename=None):
    if filename is None:
        filename = test_particles.__name__
    XX = [[-0.51,1.5], [1.51,1.5]]
    for i in range(-3, 5):
        for j in range(-6, 1):
            XX.append([i,j])
    XX = np.array(XX)
    dr = DRUHG(max_ranking=200, algorithm='slow', limitL=1, limitH=len(XX), verbose=False, do_edges=True)
    dr.fit(XX)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')

    s = 2*len(XX) - 2
    print(dr.mst_)
    print(dr.labels_)
    print(dr.mst_[s-1], dr.mst_[s-2], XX[dr.mst_[s-1]], XX[dr.mst_[s-2]])
    # two points are further metrically but close reciprocally
    assert (dr.mst_[s-4]*dr.mst_[s-3] == 0)
    assert (dr.mst_[s-4] + dr.mst_[s-3] == 1)
    assert _not_fail_all

def test_particles_big(filename=None):
    if filename is None:
        filename = test_particles_big.__name__
    XX = [[-0.51,1.5], [1.51,1.5]]
    for i in range(-5, 7):
        for j in range(-8, 1):
            XX.append([i,j])
    XX = np.array(XX)
    dr = DRUHG(max_ranking=200, algorithm='slow', limitL=1, limitH=len(XX), verbose=False, do_edges=True)
    dr.fit(XX)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')

    s = 2*len(XX) - 2
    print(dr.mst_)
    print(dr.labels_)
    print(dr.mst_[s-1], dr.mst_[s-2], XX[dr.mst_[s-1]], XX[dr.mst_[s-2]])
    # two points are further metrically but close reciprocally
    assert (dr.mst_[s-4]*dr.mst_[s-3] == 0)
    assert (dr.mst_[s-4] + dr.mst_[s-3] == 1)
    assert _not_fail_all

#
def test_bomb():
    XX = [[0.,1.],[0.,2.],[0.,3.],[0.,4.],[0.,5.]]
    for i in range(-3, 4):
        for j in range(-6, 1):
            XX.append([i,j])
    XX = np.array(XX)
    dr = DRUHG(algorithm='slow', max_ranking=200, limitL=1, limitH=len(XX), verbose=False, do_edges=True)
    dr.fit(XX)
    s = 2*len(XX) - 2
    print(dr.mst_)
    x = 12
    labs = dr.labels_
    # fuse is separate
    print(labs)
    assert (labs[0]==labs[1]==labs[2]==labs[3])
    assert (np.count_nonzero(labs == labs[0]) == 4)
    assert _not_fail_all

def test_t(filename=None):
    if filename is None:
        filename = test_t.__name__
    XX = []
    for i in range(1, 10):
        XX.append([0.,i])
    for j in range(-9, 10):
        XX.append([j,0.])
    XX = np.array(XX)
    # np.random.shuffle(XX)
    dr = DRUHG(max_ranking=200, algorithm='slow', verbose=False)
    dr.fit(XX)
    labels = dr.labels_
    n_clusters = len(set(labels)) - int(-1 in labels)
    print('n_clusters', n_clusters)
    print('labels', len(labels), labels)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')
    # t-center is an outlier too
    assert (n_clusters == 3)
    assert (np.count_nonzero(labels == -1) == 4)
    assert _not_fail_all

#
def test_cross(filename=None):
    if filename is None:
        filename = test_cross.__name__
    XX = []
    for j in range(-10, 10):
        XX.append([j,0.])
    for i in range(1, 10):
        XX.append([0., i])
        XX.append([0., i - 10])
    XX = np.array(XX)
    # np.random.shuffle(XX)
    dr = DRUHG(max_ranking=200, verbose=False)
    dr.fit(XX)
    print(XX)
    # center is an outlier
    labels = dr.labels_
    n_clusters = len(set(labels)) - int(-1 in labels)
    print('n_clusters', n_clusters)
    print('labels', len(labels))

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')

    assert (n_clusters == 4)
    assert (np.count_nonzero(labels == -1) == 5)
    assert _not_fail_all


def test_cube(size=5, scale=1., showplot=True, filename=None):
    if filename is None:
        filename = test_cube.__name__

    XX = []
    for i in range(0, size):
        for j in range(0, size):
            for k in range(0, size):
                XX.append([i*scale,j*scale,k*scale])
    XX = np.array(XX)
    np.random.shuffle(XX)
    print(XX)
    # for i, x in enumerate(XX):
    #     print(i, x)
    dr = DRUHG(algorithm='slow', max_ranking=200, limitL=1, limitH=int(len(XX)/2), verbose=False, do_edges=True)
    dr.fit(XX)
    s = 2*len(XX) - 2
    print(dr.mst_)
    # assert (0==1)
    labels = dr.labels_
    unique, counts = np.unique(labels, return_counts=True)
    print(unique, counts)
    # labels = dr.relabel(limitL=1, limitH=len(XX)/2)

    n_clusters = len(set(labels)) - int(-1 in labels)
    print('n_clusters', n_clusters, set(labels))
    # print('labels', labels)

    # sorteds = np.argsort(counts)
    # dr.relabel(limitL=1, limitH=len(XX), exclude=[unique[sorteds[0]], unique[sorteds[1]]])
    # labels = dr.labels_



    if _test_extra_visualisation and showplot and _plot_graph:
        import seaborn as sns

        # np.test() # helps if the plot not working
        plt.close('all')
        fig = plt.figure()
        ax = Axes3D(fig)

        unique, counts = np.unique(labels, return_counts=True)
        sorteds = np.argsort(counts)
        s = len(sorteds)

        i = sorteds[s - 1]
        max_size = counts[i]
        if unique[i] == 0:
            max_size = counts[sorteds[s - 2]]

        color_map = {}
        palette = sns.color_palette('bright', s + 1)
        col = 0
        a = (1. - 0.3) / (max_size - 1)
        b = 0.3 - a
        while s:
            s -= 1
            i = sorteds[s]
            if unique[i] == 0:
                continue
            alpha = a * counts[i] + b
            color_map[unique[i]] = palette[col] + (alpha,)
            col += 1

        color_map[0] = (0., 0., 0., 0.15)
        colors = [color_map[x] for x in labels]
        s = [50 for x in labels]
        # ax = fig.add_subplot(111, projection='3d')
        ax.scatter(XX[:, 0:1], XX[:, 1:2], XX[:, 2:3], c=colors, s=s)

        for i in range(0,len(labels)-1):
            print(i, XX[dr.mst_[2*i]], colors[dr.mst_[2*i]])
            x1, y1, z1 = list(XX[dr.mst_[2 * i]])
            x2, y2, z2 = list(XX[dr.mst_[2 * i + 1]])

            ax.plot([x1,x2], [y1,y2], [z1,z2], c=colors[dr.mst_[2*i]])
        # plt.pause(10)
        plt.show()
        plt.savefig(filename+'1'+'.png')

    if showplot and _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')

    assert (n_clusters == 1+6+12)
    unique, counts = np.unique(labels, return_counts=True)

    assert (-1 in unique)
    i = np.where(unique == -1)[0][0]

    assert(counts[i] == 8)
    counts = np.delete(counts,[i])

    counts = np.sort(counts)[::-1]
    print(counts, unique)

    assert ((size-2)**3==counts[0])

    for i in range(1, 1+6):
       assert (counts[i]==(size-2)**2)


    for i in range(7, 1+6+12):
        assert (counts[i] == size-2)

    # # assert (False)
    # # labels = dr.relabel(limitL=1)
    # labels = dr.relabel(limitL=1, limitH=len(XX))
    # print('out')
    # n_clusters = len(set(labels)) - int(-1 in labels)
    # print('n_clusters2', n_clusters, set(labels))
    # # print('labels2', labels)
    # assert (n_clusters == 1+6+12)

    assert _not_fail_all

def test_loop_cube():
    return
    k = 100
    while k!=0:
        # print('+++++++++++++PREVED+++++++++++++', k - 1000)
        test_cube(showplot=False)
        # assert (False)
        k-=1

def test_loop_square():
    return
    k = 100
    while k!=0:
        # print('+++++++++++++PREVED+++++++++++++', k - 1000)
        test_square(False)
        # assert (False)
        k-=1
    assert _not_fail_all

def test_druhg_sparse():
    row = np.array([0, 0, 1, 2, 2, 2])
    col = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6])
    sparse_X = sparse.csr_matrix((data, (row, col)), shape=(3, 3))

    dr = DRUHG()
    dr.fit(sparse_X)
    print('sparse labels', dr.labels_)
    assert _not_fail_all

def test_druhg_distance_matrix(filename=None):
    if filename is None:
        filename = test_druhg_distance_matrix.__name__
    D = distance.squareform(distance.pdist(X))
    D /= np.max(D)

    print(D.shape)
    dr = druhg(D, metric='precomputed')
    print(dr)
    n_clusters = len(set(dr[0])) - int(-1 in dr[0])
    print(n_clusters)
    if _plot_graph:
        plt.close('all')
        dr = DRUHG(metric="precomputed").fit(D)
        dr.plot(dr.labels_)
        plt.savefig(filename+'1'+'.png')
    assert(n_clusters==4)

    dr = DRUHG(metric="precomputed", limitL=5).fit(D)
    labels = dr.labels_
    print(labels)
    n_clusters = len(set(labels)) - int(-1 in labels)
    print(n_clusters)
    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'2'+'.png')

    assert(n_clusters==4)
    assert _not_fail_all

def test_moons_and_blobs():
    XX = X
    dr = DRUHG(max_ranking=50, verbose=False)
    dr.fit(XX)
    labels = dr.labels_
    n_clusters = len(set(labels)) - int(-1 in labels)
    # expecting 4 clusters
    print(labels)

    assert (n_clusters == 4)
    assert _not_fail_all
#
def test_hdbscan_clusterable_data(filename=None):
    if filename is None:
        filename = test_hdbscan_clusterable_data.__name__
    XX = np.load('druhg\\tests\\clusterable_data.npy')
    dr = DRUHG(max_ranking=50, verbose=False)
    dr.fit(XX)
    labels = dr.labels_
    uniques, counts = np.unique(labels, True)
    print(uniques, counts)
    n_clusters = len(set(labels)) - int(-1 in labels)
    print(n_clusters)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')

    assert (n_clusters==6)
    assert _not_fail_all

def test_three_blobs(filename=None):
    if filename is None:
        filename = test_three_blobs.__name__
    XX = np.load('druhg\\tests\\three_blobs.npy')
    dr = DRUHG(max_ranking=3550, verbose=False)
    dr.fit(XX)
    labels = dr.labels_
    uniques, counts = np.unique(labels, True)
    print(uniques, counts)
    n_clusters = len(set(labels)) - int(-1 in labels)
    print(n_clusters)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')

    assert (n_clusters==3)
    assert _not_fail_all

def test_synthetic_outliers(filename=None):
    if filename is None:
        filename = test_synthetic_outliers.__name__
    XX = pd.read_csv('druhg\\tests\\synthetic.csv', sep=',')
    XX.drop(u'outlier', axis=1, inplace=True)
    XX = np.array(XX)
    dr = DRUHG(algorithm='slow', max_ranking=200, limitH=len(XX), exclude=[1978,1973], verbose=False)
    dr.fit(XX)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')

    # values, counts = np.unique(dr.labels_, return_counts=True)
    # for i, v in enumerate(values):
    #     print(v, counts[i])

    labels = dr.labels_
    # labels = dr.relabel(limitL=1)
    n_clusters = len(set(labels)) - int(-1 in labels)
    print(labels)
    print('n_clusters', n_clusters)
    assert (n_clusters==6)
    assert _not_fail_all


def test_compound1(filename=None):
    if filename is None:
        filename = test_compound1.__name__
    XX = pd.read_csv('druhg/tests/Compound.csv', sep=',', header=None).drop(2, axis=1)
    XX = np.array(XX)
    dr = DRUHG(max_ranking=1550, limitL=3, limitH=len(XX), verbose=False)
    dr.fit(XX)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'1'+'.png')

    labels = dr.labels_
    # labels = dr.relabel(limitL=1)
    n_clusters = len(set(labels)) - int(-1 in labels)
    # np.save('labels_compound', labels)
    print('n_clusters', n_clusters, set(labels))
    exc = labels[398]
    # dr.relabel(limitL=3, limitH=len(XX), exclude=[exc])
    # labels = dr.labels_
    n_clusters2 = len(set(labels)) - int(-1 in labels)
    print(exc, n_clusters2, set(labels))

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'2'+'.png')

    exc2 = labels[398]
    dr.relabel(limitL=3, limitH=len(XX), exclude=[exc, exc2])
    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'3'+'.png')


    assert (n_clusters==4)
    assert _not_fail_all


def test_compound_egg(filename=None):
    if filename is None:
        filename = test_compound_egg.__name__
    XX = pd.read_csv('druhg/tests/Compound.csv', sep=',', header=None)
    XX = XX[(XX[2]==6) | (XX[2]==5)]
    XX = XX.drop(2, axis=1)
    XX = np.array(XX)
    dr = DRUHG(max_ranking=2000, limitL=3, limitH=len(XX), verbose=False)
    dr.fit(XX)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'.png')

    labels = dr.labels_
    # labels = dr.relabel(limitL=1)
    n_clusters = len(set(labels)) - int(-1 in labels)
    # np.save('labels_compound', labels)
    print('n_clusters', n_clusters, set(labels))
    # dr.relabel(limitL=3, limitH=len(XX), exclude=[exc])
    # labels = dr.labels_
    n_clusters2 = len(set(labels)) - int(-1 in labels)
    print( n_clusters2, set(labels))

    assert (n_clusters==2)
    assert _not_fail_all

def test_copycat():
    XX = [[0]]*100
    XX = np.array(XX)
    dr = DRUHG(max_ranking=200, limitL=1, verbose=False, do_edges=True)
    dr.fit(XX)
    print(dr.mst_[0], dr.mst_[1])
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    labels = dr.labels_
    assert (all(x == labels[0] for x in labels))
    assert _not_fail_all


def test_copycats(): # should fail until weights are made
    XX = np.concatenate( ([[0]]*100, [[1]]*100))
    dr = DRUHG(max_ranking=10, limitL=1, verbose=False, do_edges=True)
    dr.fit(XX)
    print(dr.mst_[0], dr.mst_[1])
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    labels = dr.labels_
    assert (not all(x == labels[0] for x in labels))
    assert (labels[0] != labels[-1])

    uniques, counts = np.unique(labels, True)
    print(uniques, counts)
    n_clusters = len(set(labels)) - int(-1 in labels)
    assert (n_clusters==2)
    assert _not_fail_all


def test_copycats2():
    XX = np.concatenate( ([[0]]*10, [[1]]*10))
    dr = DRUHG(max_ranking=200, limitL=1, verbose=False, do_edges=True)
    dr.fit(XX)
    print(dr.mst_[0], dr.mst_[1])
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    labels = dr.labels_
    assert (not all(x == labels[0] for x in labels))
    assert (labels[0] != labels[-1])

    uniques, counts = np.unique(labels, True)
    print(uniques, counts)
    n_clusters = len(set(labels)) - int(-1 in labels)
    assert (n_clusters==2)
    assert _not_fail_all


def test_copycats3(): # should fail until weights are made
    XX = np.concatenate( ([[0]]*100, [[1]]*100, [[2]]*5) )
    dr = DRUHG(max_ranking=10, limitL=1, limitH=250, verbose=False, do_edges=True)
    dr.fit(XX)
    print(dr.mst_[0], dr.mst_[1])
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    labels = dr.labels_
    assert (not all(x == labels[0] for x in labels))
    assert (labels[0] != labels[-1])

    uniques, counts = np.unique(labels, True)
    print(uniques, counts)
    n_clusters = len(set(labels)) - int(-1 in labels)
    assert (n_clusters==3)
    assert _not_fail_all

#
# def test_speed0():
#     XX = []
#     size = 5
#     for i in range(0, size):
#         for j in range(0, size):
#             for k in range(0, size):
#                 XX.append([i, j, k])
#
#     XX = pd.read_csv('druhg\\tests\\synthetic.csv', sep=',')
#     XX.drop(u'outlier', axis=1, inplace=True)
#
#     return np.array(XX)
#
#
# def test_speed1(k=100):
#     XX = test_speed0()
#     while k!=0:
#         k -= 1
#         np.random.shuffle(XX)
#         dr = DRUHG(algorithm='fast', max_ranking=200, limitL=1, limitH=int(len(XX) / 2), verbose=False)
#         dr.fit(XX)
#
# def test_speed2(k=100):
#     XX = test_speed0()
#     while k!=0:
#         k -= 1
#         np.random.shuffle(XX)
#         dr = DRUHG(algorithm='fastminus', max_ranking=200, limitL=1, limitH=int(len(XX) / 2), verbose=False)
#         dr.fit(XX)
#
# def test_speed3(k=100):
#     XX = test_speed0()
#     while k!=0:
#         k -= 1
#         np.random.shuffle(XX)
#         dr = DRUHG(algorithm='slow+', max_ranking=200, limitL=1, limitH=int(len(XX) / 2), verbose=False)
#         dr.fit(XX)
#
# def test_speed4(k=100):
#     XX = test_speed0()
#     while k!=0:
#         k -= 1
#         np.random.shuffle(XX)
#         dr = DRUHG(algorithm='slow', max_ranking=200, limitL=1, limitH=int(len(XX) / 2), verbose=False)
#         dr.fit(XX)
#
# def test_speed5(k=100):
#     XX = test_speed0()
#     while k!=0:
#         k -= 1
#         np.random.shuffle(XX)
#         dr = DRUHG(algorithm='slow', max_ranking=200, limitL=1, limitH=int(len(XX) / 2), verbose=False)
#         dr.fit(XX)



def test_triangle(scale = 1., height1 = 0.45, height2 = 0.5, dis=0,  filename = None, three_clusters = True):
    if filename is None:
        filename = test_triangle.__name__
    XX = [[scale*height1, 0.], [scale*height2, 0], [0, scale*dis]]
    XX = np.array(XX)
    dr = DRUHG(algorithm='slow', max_ranking=200, limitL=1, limitH=1000, verbose=False, do_edges=True)
    dr.fit(XX)

    print(XX)

    labels = dr.labels_
    print('pairs', dr.mst_)
    print('labels', dr.labels_)
    n_clusters = len(set(labels)) - int(-1 in labels)
    print('n_clusters', n_clusters)
    return not all(x == labels[0] for x in labels)


def test_triangle_full(filename = None):
    full_res = []
    res1 = []
    steps = 31
    for i in range(0, steps):
        dis = i*0.1

        if test_triangle(scale=1., dis=dis,
                         height1=0.57735026918962573, height2=-0.57735026918962573,
                         filename="test_triangle_equilateral_"+str(i)):
            res1.append(True)
        else:
            res1.append(dis)
        print("{:.4f}".format(dis), '======equilateral======')
    print('result equilateral', res1)

    res2 = []
    for i in range(0, steps):
        dis = i*0.1
        if test_triangle(scale=1., dis=dis,
                                 height1=1., height2=0,
                                 filename="test_triangle_right"+str(i)):
            res2.append(True)
        else:
            res2.append(dis)
        print("{:.4f}".format(dis), '======right======')
    print('result right', res2)
    print('result equilateral', res1)

    assert (True in res1)
    assert (True in res2)
    assert (not all(x for x in res1))
    assert (not all(x for x in res2))

    assert _not_fail_all

def test_triangle_line(filename = None):
    res = []
    for i in range(0, 17):
        h = 0.45+i*0.01
        res.append(test_triangle(scale=1., height=h, filename="test_triangle_"+str(i)))
        print(h, '==========================')
    print(res)
    assert (False in res)
    assert (True in res)
    assert _not_fail_all

def test_run(filename=None):
    if filename is None:
        filename = test_run.__name__
    XX = pd.read_csv('druhg\\tests\\chameleon.csv', sep='\t', header=None)
    XX = np.array(XX)
    plt.style.use('dark_background')
    dr = DRUHG(max_ranking=4200, limitL=1, limitH=len(XX)/4, do_edges=True)
    dr.fit(XX)
    if _plot_graph:
        plt.close('all')
        dr.plot(core_color='brown')
        plt.savefig(filename+'1'+'.png')
    assert (False)

def test_chameleon(filename=None):
    if filename is None:
        filename = test_chameleon.__name__
    XX = pd.read_csv('druhg\\tests\\chameleon.csv', sep='\t', header=None)
    XX = np.array(XX)
    # using time module

    print('start', time.time())
    dr = DRUHG(max_ranking=4200, limitL=1, limitH=len(XX), verbose=False, algorithm='slow')
    print('init', time.time())
    dr.fit(XX)
    print('afterfit',time.time())
    labels = dr.labels_
    # labels = dr.relabel(limitL=1)
    values, counts = np.unique(labels, return_counts=True)
    n_clusters = 0
    for i, c in enumerate(counts):
        print(i, c, values[i])
        if c > 500 and values[i] >= 0:
            n_clusters += 1
    print('n_clusters', n_clusters)

    print('preplot',time.time())
    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'1'+'.png')
    # assert (False)
    print('afterplot',time.time())

    dr = DRUHG(max_ranking=1200, limitL=1, limitH=int(len(XX)/4), exclude=[], verbose=False, algorithm='slow')
    dr.fit(XX)

    values, counts = np.unique(dr.labels_, return_counts=True)
    n_clusters2 = 0
    for i, c in enumerate(counts):
        print(i, c, values[i])
        if c > 500 and values[i] >= 0:
            n_clusters2 += 1
    print('n_clusters2', n_clusters2)

    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'2'+'.png')

    if _plot_graph and _test_extra_visualisation:
        plt.close('all')
        dr.single_linkage_.plot(dr.labels_)
        plt.savefig(filename +'plot'+ '.png')

    exc = dr.labels_[3024]
    dr.relabel(limitL=3, limitH=int(len(XX)/4), exclude=[exc])
    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'3'+'.png')

    exc = dr.labels_[3024]
    dr.relabel(limitL=3, limitH=int(len(XX)/8), exclude=[exc])
    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename+'4'+'.png')

    assert (n_clusters2==6)
    assert (n_clusters==6)
    assert _not_fail_all


# def test_boba(filename=None):
#     if filename is None:
#         filename = test_boba.__name__
#     XX = pd.read_csv('druhg\\tests\\chameleon.csv', sep='\t', header=None)
#     XX = np.array(XX)
#     lim = 1.
#     while lim >= 0:
#
#         dr = DRUHG(max_ranking=4200, limitL = 1, limitH=lim, verbose=False, algorithm='slow')
#         dr.fit(XX)
#         if _plot_graph:
#             plt.close('all')
#             dr.plot(dr.labels_)
#             plt.savefig(filename+str(lim)+'.png')
#         lim -= 0.05

def test_amt(filename=None):
    if filename is None:
        filename = test_amt.__name__
    XX = pd.read_csv('druhg\\tests\\chameleon.csv', sep='\t', header=None)
    XX = np.array(XX)
    dr = DRUHG(max_ranking=4200, limitL=1, limitH=0.25, verbose=False, algorithm='slow')
    dr.fit(XX)
    if _plot_graph:
        plt.close('all')
        dr.plot(dr.labels_)
        plt.savefig(filename + '.png')
    assert _not_fail_all



"""
Tests for DRUHG clustering algorithm
"""
import pickle
import numpy as np
import pandas as pd
from scipy.spatial import distance
from scipy import sparse
from scipy import stats
import pytest  #delete __init__.py or it wont work


from druhg import (DRUHG,
                   druhg)

from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# from sklearn.cluster.tests.common import generate_clustered_data
import sklearn.datasets as datasets
from sklearn.datasets import make_blobs
from sklearn.utils import shuffle
from sklearn.preprocessing import StandardScaler
from scipy.stats import mode
from sklearn.metrics import adjusted_rand_score

from tempfile import mkdtemp
from functools import wraps

import warnings

_plot_graph = True
_not_fail_all = False

blobs, _ = datasets.make_blobs(n_samples=30, centers=[(-1., -1.), (1.0, 1.0)], cluster_std=0.25)
X = np.vstack([blobs])

X = [[ 0.92106143 , 0.90547575],
 [ 0.77740108,  0.74381424],
 [-0.54707238, -0.93632275],
 [ 1.14528332,  0.47395599],
 [-1.18154265, -0.86885681],
 [-0.99693863, -1.534588  ],
 [-0.6993809,  -1.18147767],
 [ 1.28328471,  0.77446717],
 [-0.69767129, -0.70035315],
 [ 1.0087693 ,  1.0576548 ],
 [ 0.6074741 ,  1.37124275],
 [ 1.27818319,  1.33145458],
 [ 1.16809938,  1.05568675],
 [-0.96391345, -0.76431969],
 [-0.97849489, -0.64393941],
 [-0.76445103, -0.92221768],
 [-0.66520734, -1.33962011],
 [-0.82256141, -1.23753227],
 [ 0.96583008,  0.87435429],
 [-1.46553152, -0.96775917],
 [-0.7730766 , -1.1289692 ],
 [ 1.37473563,  0.99939175],
 [ 1.45518458,  0.75611808],
 [ 1.145298  ,  0.91818625],
 [-1.04021956, -1.02153293],
 [ 1.23676014,  1.45337014],
 [ 1.02380103,  0.91979054],
 [-1.13063934, -1.43201801],
 [ 0.9759653 ,  1.19603341],
 [-0.82044957, -1.20135226]]

def test_splashes(filename=None):
    if filename is None:
        filename = test_splashes.__name__
    XX = np.array(X)
    for i in range(1, 15):
        print('HELLO', i)
        print(XX)
        dr = DRUHG(max_ranking=60, verbose=False)
        dr.fit(XX)
        dr.develop(XX)
        print('labels_', dr.labels_)

        if _plot_graph:
            plt.close('all')
            axis = dr.minimum_spanning_tree_.plot()
            axis.axhline(0, color='gray')
            axis.axvline(0, color='gray')
            axis.set_xlim(-20.0, 20.0)
            axis.set_ylim(-20.0, 20.0)
            plt.savefig(filename+str(i)+'.png')
        print('dr.new_data_', dr.new_data_)
        XX = np.array(dr.new_data_)


    assert _not_fail_all
